// Copyright 2021 Northern.tech AS
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

// Code generated by mockery v2.2.2. DO NOT EDIT.

package mocks

import (
	mock "github.com/stretchr/testify/mock"

	ws "github.com/mendersoftware/go-lib-micro/ws"
)

// ResponseWriter is an autogenerated mock type for the ResponseWriter type
type ResponseWriter struct {
	mock.Mock
}

// WriteProtoMsg provides a mock function with given fields: msg
func (_m *ResponseWriter) WriteProtoMsg(msg *ws.ProtoMsg) error {
	ret := _m.Called(msg)

	var r0 error
	if rf, ok := ret.Get(0).(func(*ws.ProtoMsg) error); ok {
		r0 = rf(msg)
	} else {
		r0 = ret.Error(0)
	}

	return r0
}
