// Copyright 2021 Northern.tech AS
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
package procps

import (
	"os/exec"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

func TestMenderShellProcPs(t *testing.T) {
	cmd := exec.Command("sleep", "16")
	err := cmd.Start()
	assert.NoError(t, err)

	assert.True(t, ProcessExists(cmd.Process.Pid))

	err = TerminateAndWait(cmd.Process.Pid, cmd, time.Second)
	assert.Nil(t, err)
	assert.False(t, ProcessExists(cmd.Process.Pid))
}
